<?php

namespace Drupal\mustache\Element;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Render\Element\RenderElement;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Url;
use Drupal\mustache\Exception\MustacheUrlNotFoundException;
use Drupal\mustache\Render\Markup;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Provides an element for rendering Mustache templates.
 *
 * @RenderElement("mustache")
 */
class Mustache extends RenderElement {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#theme' => 'mustache',
      '#pre_render' => [
        [$class, 'generateContentMarkup'],
      ],
    ];
  }

  /**
   * Generates content markup for a Mustache template.
   *
   * @param array $element
   *   The element to render.
   *
   * @return array
   *   The element to render.
   *
   * @throws \Exception
   *   In case something went wrong on placeholder rendering,
   *   or when the template or data url could not be found.
   */
  public static function generateContentMarkup(array $element) {
    if (!isset($element['#cache'])) {
      // Caching will be bubbled up if necessary.
      $element['#cache'] = [];
    }

    /** @var \Drupal\mustache\MustacheTemplates $templates */
    $templates = \Drupal::service('mustache.templates');
    $template = $templates->getContent($element['#template']);

    $data = isset($element['#data']) ? $element['#data'] : NULL;
    $select = isset($element['#select']) ? $element['#select'] : NULL;
    if (is_string($select)) {
      $select = [$select];
    }

    // Obtain the Json data url, if given.
    $url = NULL;
    if ($data instanceof Url) {
      // The #data value represents a url object to fetch data from.
      $url = $data;
      $data = NULL;
    }
    elseif (is_string($data)) {
      // The #data value represents a url string to fetch data from.
      try {
        $url = Url::fromUri($data);
      }
      catch (\InvalidArgumentException $e) {
        $url = Url::fromUserInput($data);
      }
      $data = NULL;
    }
    if (isset($url)) {
      $url = $url->isExternal() ? $url->setAbsolute()->toString() : $url->setAbsolute(FALSE)->toString();
    }

    if (!empty($element['#sync'])) {
      $sync_url = $url;
      if (!empty($element['#sync']['data_url'])) {
        $sync_url = $element['#sync']['data_url'];
        try {
          $sync_url = ($sync_url instanceof Url) ? $sync_url : Url::fromUri($sync_url);
        }
        catch (\InvalidArgumentException $e) {
          $sync_url = Url::fromUserInput($sync_url);
        }
        $sync_url = $sync_url->isExternal() ? $sync_url->setAbsolute() : $sync_url->setAbsolute(FALSE);
      }
      if (!isset($sync_url)) {
        throw new MustacheUrlNotFoundException(t('Could not find Json data source url for rendering template @template via Mustache.js.', ['@template' => $element['#template']]));
      }
      $element['#use_sync'] = TRUE;
      $element['#sync']['data_url'] = $sync_url;

      $sync_select = $select;
      if (!empty($element['#sync']['select'])) {
        $sync_select = $element['#sync']['select'];
        if (is_string($sync_select)) {
          $sync_select = [$sync_select];
        }
      }
      $element['#sync']['select'] = json_encode($sync_select);

      $template_encoded = trim(substr(json_encode($template), 1, -1));
      $element['#sync']['template'] = $template_encoded;

      $element_id = 'msync-' . Crypt::randomBytesBase64(4);
      $element['#sync']['attributes'] = new Attribute(['id' => $element_id, 'class' => ['mustache-sync', 'not-synced']]);
      $element['#sync']['wrapper_tag'] = !empty($element['#sync']['wrapper_tag']) ? $element['#sync']['wrapper_tag'] : 'div';
      $element['#attached']['library'][] = 'mustache/sync';
      // Use a period fallback value if necessary.
      if (empty($element['#sync']['period'])) {
        $element['#sync']['period'] = 0;
      }
    }
    else {
      $element['#use_sync'] = FALSE;
      $element['#sync'] = [];
    }

    if (!empty($element['#placeholder'])) {
      // Instead of rendering the content via Mustache.php,
      // render an arbitrary placeholder, by respecting its cache metadata.
      $element_metadata = CacheableMetadata::createFromRenderArray($element);
      $placeholder_metadata = CacheableMetadata::createFromRenderArray($element['#placeholder']);
      $element_metadata->merge($placeholder_metadata)->applyTo($element);

      /** @var \Drupal\Core\Render\Renderer $renderer */
      $renderer = \Drupal::service('renderer');
      $element['#content'] = $renderer->render($element['#placeholder']);
    }
    else {
      // Render the content via Mustache.php.
      if (!isset($data) && isset($url)) {
        // Rendering via Mustache.php requires some data,
        // which now must be fetched from the url.
        $client = \Drupal::httpClient();
        try {
          $response = $client->request('GET', $url);
        }
        catch (GuzzleException $e) {
          \Drupal::logger('mustache')->error(t('Failed to fetch data from @url. Exception message: @message', ['@url' => $url, '@message' => $e->getMessage()]));
          return static::emptyElement();
        }

        if (!($response->getStatusCode() == 200)) {
          \Drupal::logger('mustache')->error(t('Failed to fetch data from @url. Returned status code was: @status', ['@url' => $url, '@status' => $response->getStatusCode()]));
          return static::emptyElement();
        }

        $body = $response->getBody()->getContents();
        $data = json_decode($body, TRUE);
        if (NULL === $data) {
          \Drupal::logger('mustache')->error(t('Invalid Json received from from @url.', ['@url' => $url]));
          return static::emptyElement();
        }
      }
      if (!isset($data)) {
        // In case we still don't have any data,
        // we cannot continue rendering.
        \Drupal::logger('mustache')->error(t('Failed to get data for rendering Mustache template @template', ['@template' => $element['#template']]));
        return static::emptyElement();
      }

      if (!empty($select) && !empty($data)) {
        if (!is_array($data)) {
          // In case something different was passed,
          // try to convert it. When it fails, it should fail hard.
          $data = (array) $data;
        }
        $subset_exists = NULL;
        $data = NestedArray::getValue($data, $select, $subset_exists);
        if (!$subset_exists) {
          // Subset must be an array.
          $data = [];
        }
      }

      /** @var \Mustache_Engine $mustache */
      $mustache = \Drupal::service('mustache.engine');
      $output = $mustache->loadTemplate($template)->render($data);
      $element['#content'] = Markup::create($output);
    }

    return $element;
  }

  /**
   * Returns an empty element which is not cacheable.
   *
   * This element would only be used at unusual circumstances.
   * Because it's not what the result should look like,
   * the caching is completely disabled for being able
   * to generate an expected result as soon as possible.
   *
   * @return array
   *   The empty element.
   */
  public static function emptyElement() {
    return ['#cache' => ['max-age' => 0], '#printed' => TRUE, '#markup' => Markup::create('')];
  }

}
