<?php

namespace Drupal\mustache;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Theme\ActiveTheme;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\mustache\Exception\MustacheTemplateNotFoundException;

/**
 * The finder of Mustache templates.
 */
class MustacheTemplates {

  /**
   * A list of known template file paths.
   *
   * @var string[]
   */
  protected $templates;

  /**
   * A collection of already scanned files.
   *
   * @var array
   */
  protected $scanned;

  /**
   * In-memory hold of the last template content.
   *
   * @var string[]
   */
  protected $cachedTemplateContent = [];

  /**
   * Template locations defined by modules.
   *
   * @var array
   */
  protected $templateInfo;

  /**
   * The cache backend to store template path information.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The theme manager.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected $themeManager;

  /**
   * MustacheTemplates constructor.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache backend to store template path information.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Theme\ThemeManagerInterface $theme_manager
   *   The theme manager.
   */
  public function __construct(CacheBackendInterface $cache, ModuleHandlerInterface $module_handler, ThemeManagerInterface $theme_manager) {
    $this->cache = $cache;
    $this->moduleHandler = $module_handler;
    $this->themeManager = $theme_manager;
  }

  /**
   * Get the file content for the given template name.
   *
   * @param string $name
   *   The name of the template, without file ending.
   *
   * @return string
   *   The content of the template.
   *
   * @throws \Drupal\mustache\Exception\MustacheTemplateNotFoundException
   *   In case the template file could not be found.
   */
  public function getContent($name) {
    if (isset($this->cachedTemplateContent[$name])) {
      return $this->cachedTemplateContent[$name];
    }
    if (count($this->cachedTemplateContent) > 3) {
      // Just cache a limited set of file contents.
      $this->cachedTemplateContent = [];
    }
    $content = file_get_contents($this->find($name));
    $this->cachedTemplateContent[$name] = $content;
    return $content;
  }

  /**
   * Looks up the file location for the given template name.
   *
   * @param string $name
   *   The name of the template, without file ending.
   *
   * @return string
   *   The file path as string if found.
   *
   * @throws \Drupal\mustache\Exception\MustacheTemplateNotFoundException
   *   In case the template file could not be found.
   */
  public function find($name) {
    $active_theme = $this->themeManager->getActiveTheme();
    $cid = 'mustache:' . $active_theme->getName() . ':' . $name;
    if (isset($this->templates[$cid])) {
      return $this->templates[$cid];
    }

    if ($cached = $this->cache->get($cid)) {
      if (file_exists($cached->data)) {
        $this->templates[$cid] = $cached->data;
        return $this->templates[$cid];
      }
    }

    $template_file = NULL;

    $template_file = $this->findTemplateInTheme($active_theme, $name);

    if (!isset($template_file)) {
      foreach ($active_theme->getBaseThemes() as $base_theme) {
        if ($template_file = $this->findTemplateInTheme($base_theme, $name)) {
          break;
        }
      }
    }

    if (!isset($template_file)) {
      if (!isset($this->templateInfo)) {
        $this->templateInfo = $this->moduleHandler->invokeAll('mustache_template');
        $this->moduleHandler->alter('mustache_template', $this->templateInfo);
        if (isset($this->templateInfo[$name]['file'])) {
          $template_file = $this->templateInfo[$name]['file'];
        }
      }
    }

    if (!isset($template_file)) {
      throw new MustacheTemplateNotFoundException(t('Mustache template not found for name @name.', ['@name' => $name]));
    }
    elseif (!file_exists($template_file)) {
      throw new MustacheTemplateNotFoundException(t('The registered file @file for the Mustache template @name could not be found.', ['@file' => $template_file, '@name' => $name]));
    }

    $this->templates[$cid] = $template_file;
    $this->cache->set($cid, $template_file);
    return $template_file;
  }

  /**
   * Try to find the template file for the name inside the given theme.
   *
   * @param ActiveTheme $theme
   *   The theme object.
   * @param $name
   *   The template name, without file ending.
   *
   * @return string|null
   *   The file uri if found, NULL otherwise.
   */
  protected function findTemplateInTheme(ActiveTheme $theme, $name) {
    $theme_name = $theme->getName();
    $filename = $name . '.mustache';
    $filename_hyphen = str_replace('_', '-', $name) . '.mustache';

    if (!isset($this->scanned[$theme_name])) {
      $this->scanned[$theme_name] = file_scan_directory($theme->getPath() . '/templates', '/\.mustache\.tpl$/');
    }
    foreach ($this->scanned[$theme_name] as $scanned) {
      if (($scanned->name == $filename) || ($scanned->name == $filename_hyphen)) {
        return $scanned->uri;
      }
    }

    return NULL;
  }

}
